# -*- coding: utf-8 -*-

""" tests.test_root

Some integration tests tests for conda-content-trust that focus on
generation, signing, and verification of root metadata.  This tests GPG
integration via securesystemslib if securesystemslib can be successfully
imported.

IN CONTINUOUS INTEGRATION, this set of tests should be run with and without
securesystemslib and GPG available on the system.

Run the tests this way:
    pytest tests/test_root.py

"""

# Python2 Compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import json

import pytest

# securesystemslib is an optional dependency, and required only for signing
# root metadata via GPG.  Verification of those signatures, and signing other
# metadata with raw ed25519 signatures, does not require securesystemslib.
try:
    import securesystemslib.gpg.functions as gpg_funcs
    import securesystemslib.formats
    SSLIB_AVAILABLE = True
except ImportError:
    SSLIB_AVAILABLE = False

import conda_content_trust.common as common
import conda_content_trust.metadata_construction as metadata_construction
import conda_content_trust.signing as signing
import conda_content_trust.root_signing as root_signing
import conda_content_trust.authentication as authentication

# Note that changing these sample values breaks the sample signature, so you'd
# have to generate a new one.

# A 40-hex-character GPG public key fingerprint
SAMPLE_FINGERPRINT = '917adb684e2e9fb5ed4e59909ddd19a1268b62d0'
SAMPLE_UNKNOWN_FINGERPRINT = '0123456789abcdef0123456789abcdef01234567'

# The real key value of the public key (q, 32-byte ed25519 public key val),
# as a length-64 hex string.
SAMPLE_KEYVAL = 'c8bd83b3bfc991face417d97b9c0db011b5d256476b602b92fec92849fc2b36c'

SAMPLE_GPG_KEY_OBJ = {
  'creation_time': 1571411344,
  'hashes': ['pgp+SHA2'],
  'keyid': SAMPLE_FINGERPRINT,
  'keyval': {
    'private': '',
    'public': {'q': SAMPLE_KEYVAL}
  },
  'method': 'pgp+eddsa-ed25519',
  'type': 'eddsa'
}

SAMPLE_ROOT_MD_CONTENT = {
  'delegations': {
    'key_mgr': {'pubkeys': [], 'threshold': 1},
    'root': {
      'pubkeys': ['c8bd83b3bfc991face417d97b9c0db011b5d256476b602b92fec92849fc2b36c'],
      'threshold': 1}
  },
  'expiration': '2030-12-09T17:20:19Z',
  'metadata_spec_version': '0.6.0',
  'type': 'root',
  'version': 1
}

# To generate a new signature after changing SAMPLE_ROOT_MD_CONTENT:
# >  serialized = common.canonserialize(SAMPLE_ROOT_MD_CONTENT)
# >  new_sig = root_signing.sign_via_gpg(serialized, SAMPLE_FINGERPRINT)
# If desired, you can add:  new_sig['see_also'] = SAMPLE_FINGERPRINT to keep
# the optional value listed (the OpenPGP fingerprint of the signing key).
SAMPLE_GPG_SIG = {
  'see_also': '917adb684e2e9fb5ed4e59909ddd19a1268b62d0', # optional entry
  'other_headers': '04001608001d162104917adb684e2e9fb5ed4e59909ddd19a1268b62d005025f970318',
  'signature': '41867f58064c89acb300b1b42f4d59ec52e11e6aab05cb7f651345d878ee06d3a4f32411646134e112a7d8adc1d1304f63fb918b57cac449baba36ef0a1fbe07'}

SAMPLE_SIGNED_ROOT_MD = {
  'signatures': {
    SAMPLE_KEYVAL: SAMPLE_GPG_SIG},
  'signed': SAMPLE_ROOT_MD_CONTENT}

def test_gpg_key_retrieval_with_unknown_fingerprint():
    if not SSLIB_AVAILABLE:
        pytest.skip(
                '--TEST SKIPPED⚠️ : Unable to use GPG key retrieval or '
                'signing without securesystemslib and GPG.')
        return

    # TODO✅: Adjust this to use whatever assertRaises() functionality the
    #         testing suite we're using provides.


    with pytest.raises(securesystemslib.gpg.exceptions.KeyNotFoundError):
        full_gpg_pubkey = gpg_funcs.export_pubkey(SAMPLE_UNKNOWN_FINGERPRINT)

    print(
            '--TEST SUCCESS✅: detection of error when we pass an unknown '
            'key fingerprint to GPG for retrieval of the full public key.')



def test_gpg_signing_with_unknown_fingerprint():
    if not SSLIB_AVAILABLE:
        pytest.skip(
                '--TEST SKIPPED⚠️ : Unable to use GPG key retrieval or '
                'signing without securesystemslib and GPG.')
        return

    # TODO✅: Adjust this to use whatever assertRaises() functionality the
    #         testing suite we're using provides.
    try:
        gpg_sig = root_signing.sign_via_gpg(
                b'1234', SAMPLE_UNKNOWN_FINGERPRINT)
    except securesystemslib.gpg.exceptions.CommandError as e:
        # TODO✅: This is a clumsy check.  It's a shame we don't get better
        #         than CommandError(), but this will do for now.
        assert 'signing failed: No secret key' in e.args[0]
    else:
        assert False, 'Expected CommandError was not raised!'

    print(
            '--TEST SUCCESS✅: detection of error when we pass an unknown '
            'key fingerprint to GPG for signing.')


# def test_gpg_verification_compared_to_ssls():





def test_root_gen_sign_verify():
    # Integration test

    # Build a basic root metadata file with empty key_mgr delegation and one
    # root key, threshold 1, version 1.
    rmd = metadata_construction.build_root_metadata(
            root_version=1,
            root_pubkeys=[SAMPLE_KEYVAL], root_threshold=1,
            key_mgr_pubkeys=[], key_mgr_threshold=1)
    rmd = signing.wrap_as_signable(rmd)

    signed_portion = rmd['signed']

    canonical_signed_portion = common.canonserialize(signed_portion)


    if not SSLIB_AVAILABLE:
        pytest.skip(
                '--TEST SKIPPED⚠️ : Unable to perform GPG signing without '
                'securesystemslib and GPG.')
        return

    # gpg_key_obj = securesystemslib.gpg.functions.export_pubkey(
    #         SAMPLE_FINGERPRINT)

    gpg_sig = root_signing.sign_via_gpg(
            canonical_signed_portion, SAMPLE_FINGERPRINT)

    signed_rmd = copy.deepcopy(rmd)

    signed_rmd['signatures'][SAMPLE_KEYVAL] = gpg_sig



    # # Dump working files
    # with open('T_gpg_sig.json', 'wb') as fobj:
    #     fobj.write(common.canonserialize(gpg_sig))

    # with open('T_gpg_key_obj.json', 'wb') as fobj:
    #     fobj.write(common.canonserialize(gpg_key_obj))

    # with open('T_canonical_sigless_md.json', 'wb') as fobj:
    #     fobj.write(canonical_signed_portion)

    # with open('T_full_rmd.json', 'wb') as fobj:
    #     fobj.write(common.canonserialize(signed_rmd))



    # Verify using the SSL code and the expected pubkey object.
    # # (Purely as a test -- we wouldn't normally do this.)
    # verified = securesystemslib.gpg.functions.verify_signature(
    #     gpg_sig, gpg_key_obj, canonical_signed_portion)

    # assert verified

    authentication.verify_gpg_signature(
            gpg_sig, SAMPLE_KEYVAL, canonical_signed_portion)

    print(
            '--TEST SUCCESS✅: GPG signing (using GPG and securesystemslib) and '
            'GPG signature verification (using only cryptography)')



def test_verify_existing_root_md():

    # It's important that we are able to verify root metadata without anything
    # except old root metadata, so in particular we don't want to need the
    # full GPG public key object.  Ideally, we want only the Q value of the
    # key, but if we also need to retain the GPG key fingerprint (listed in the
    # signature itself), we can do that.....

    # with open('T_full_rmd.json', 'rb') as fobj:
    #     signed_rmd = json.load(fobj)
    #
    # gpg_sig = signed_rmd['signatures'][
    #     'bfbeb6554fca9558da7aa05c5e9952b7a1aa3995dede93f3bb89f0abecc7dc07']
    #
    # canonical_signed_portion = common.canonserialize(signed_rmd['signed'])
    #
    # with open('T_gpg_key_obj.json', 'rb') as fobj:
    #     gpg_key_obj = json.load(fobj)
    # q = gpg_key_obj['keyval']['public']['q']
    # fingerprint = gpg_key_obj['keyid']


    canonical_signed_portion = common.canonserialize(
            SAMPLE_ROOT_MD_CONTENT)

    # # First, try using securesystemslib's GPG signature verifier directly.
    # verified = securesystemslib.gpg.functions.verify_signature(
    #     SAMPLE_GPG_SIG,
    #     SAMPLE_GPG_KEY_OBJ,  # <-- We don't want conda to have to provide this.
    #     canonical_signed_portion)

    # assert verified

    # # Second, try it using my adapter, skipping a bit of ssl's process.
    # verified = root_signing.verify_gpg_sig_using_ssl(
    #         SAMPLE_GPG_SIG,
    #         SAMPLE_FINGERPRINT,
    #         SAMPLE_KEYVAL,
    #         canonical_signed_portion)

    # assert verified

    # Third, use internal code only.  (This is what we're actually going to
    # use in conda.)

    # Verify using verify_gpg_signature.
    authentication.verify_gpg_signature(
    # authentication.verify_gpg_signature(
            SAMPLE_GPG_SIG,
            SAMPLE_KEYVAL,
            canonical_signed_portion)

    print(
            '--TEST SUCCESS✅: GPG signature verification without using GPG or '
            'securesystemslib')

    # Verify using verify_signable.
    authentication.verify_signable(
            SAMPLE_SIGNED_ROOT_MD, [SAMPLE_KEYVAL], 1, gpg=True)


    # TODO ✅: Add a v2 of root to this test, and verify static v2 via v1 as
    #          well.  Also add failure modes (verifying valid v2 using v0
    #          expectations.)



# -*- coding: utf-8 -*-

""" tests.test_authentication

Unit tests for conda-content-trust/car/authentication.py
as well as integration tests for the signing.py + authentication.py.

Run the tests this way:
    pytest tests/test_authentication.py

"""

# Python2 Compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

# std libs
import copy
import os

# dependencies
import pytest
import cryptography.exceptions

# this codebase
from conda_content_trust.authentication import *
from conda_content_trust.metadata_construction import (
        gen_keys, gen_and_write_keys, # for new-key tests
        # build_repodata_verification_metadata
        )
from conda_content_trust.common import (
        PrivateKey, PublicKey, keyfiles_to_bytes, keyfiles_to_keys,
        SignatureError, MetadataVerificationError)
from conda_content_trust.signing import wrap_as_signable, sign_signable

# Some REGRESSION test data.
REG__KEYPAIR_NAME = 'keytest_old'
REG__PRIVATE_BYTES = b'\xc9\xc2\x06\r~\r\x93al&T\x84\x0bI\x83\xd0\x02!\xd8\xb6\xb6\x9c\x85\x01\x07\xdat\xb4!h\xf97'
REG__PUBLIC_BYTES = b"\x01=\xddqIb\x86m\x12\xba[\xae'?\x14\xd4\x8c\x89\xcf\x07s\xde\xe2\xdb\xf6\xd4V\x1eR\x1c\x83\xf7"
REG__PUBLIC_HEX_ROOT = 'c8bd83b3bfc991face417d97b9c0db011b5d256476b602b92fec92849fc2b36c'
REG__MESSAGE_THAT_WAS_SIGNED = b'123456\x067890'
# Signature is over REG__MESSAGE_THAT_WAS_SIGNED using key REG__PRIVATE_BYTES.
REG__SIGNATURE = b'\xb6\xda\x14\xa1\xedU\x9e\xbf\x01\xb3\xa9\x18\xc9\xb8\xbd\xccFM@\x87\x99\xe8\x98\x84C\xe4}9;\xa4\xe5\xfd\xcf\xdaau\x04\xf5\xcc\xc0\xe7O\x0f\xf0F\x91\xd3\xb8"\x7fD\x1dO)*\x1f?\xd7&\xd6\xd3\x1f\r\x0e'
REG__SIGNATURE_HEX = 'b6da14a1ed559ebf01b3a918c9b8bdcc464d408799e8988443e47d393ba4e5fdcfda617504f5ccc0e74f0ff04691d3b8227f441d4f292a1f3fd726d6d31f0d0e'
# REG__HASHED_VAL = b'string to hash\n'
# REG__HASH_HEX = '73aec9a93f4beb41a9bad14b9d1398f60e78ccefd97e4eb7d3cf26ba71dbe0ce'
# #REG__HASH_BYTES = b's\xae\xc9\xa9?K\xebA\xa9\xba\xd1K\x9d\x13\x98\xf6\x0ex\xcc\xef\xd9~N\xb7\xd3\xcf&\xbaq\xdb\xe0\xce'
# REG__REPODATA_HASHMAP = {
#     "noarch/current_repodata.json": "908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe",
#     "noarch/repodata.json": "908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe",
#     "noarch/repodata_from_packages.json": "908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe",
#     "osx-64/current_repodata.json": "8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2",
#     "osx-64/repodata.json": "8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2",
#     "osx-64/repodata_from_packages.json": "8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2"
# }
REG__TEST_TIMESTAMP = '2019-10-01T00:00:00Z'
REG__TEST_EXPIRY_DATE = '2025-01-01T10:30:00Z'
# REG__EXPECTED_UNSIGNED_REPODATA_VERIFY = {
#     'type': 'repodata_verify', 'timestamp': REG__TEST_TIMESTAMP,
#     'metadata_spec_version': '0.0.5', 'expiration': REG__TEST_EXPIRY_DATE,
#     'secured_files': {
#         'noarch/current_repodata.json': '908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe',
#         'noarch/repodata.json': '908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe',
#         'noarch/repodata_from_packages.json': '908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe',
#         'osx-64/current_repodata.json': '8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2',
#         'osx-64/repodata.json': '8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2',
#         'osx-64/repodata_from_packages.json': '8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2'}
# }
# REG__EXPECTED_REGSIGNED_REPODATA_VERIFY = {
#     # Re-sign this if its data changes: it's signed!
#     'type': 'repodata_verify', 'timestamp': '2019-10-01T00:00:00Z',
#     'metadata_spec_version': '0.0.5', 'expiration': '2025-01-01T10:30:00Z',
#     'secured_files': {
#         'noarch/current_repodata.json': '908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe',
#         'noarch/repodata.json': '908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe',
#         'noarch/repodata_from_packages.json': '908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe',
#         'osx-64/current_repodata.json': '8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2',
#         'osx-64/repodata.json': '8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2',
#         'osx-64/repodata_from_packages.json': '8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2'}
# }

TEST_ROOT_MD_V1 = {
  "signatures": {
    "c8bd83b3bfc991face417d97b9c0db011b5d256476b602b92fec92849fc2b36c": {
      "other_headers": "04001608001d162104917adb684e2e9fb5ed4e59909ddd19a1268b62d005025f96ff1a",
      "signature": "9327c2c4907c964741924420c4c35eb01805c15ec2fe5cad17bc98c0c3daf03006fcafb332eaa543a1ed212fac05f227662d8617970afc6c919ee4b78bacb004"
    }
  },
  "signed": {
    "delegations": {
      "key_mgr": {
        "pubkeys": [
          "013ddd714962866d12ba5bae273f14d48c89cf0773dee2dbf6d4561e521c83f7"
        ],
        "threshold": 1
      },
      "root": {
        "pubkeys": [
          "c8bd83b3bfc991face417d97b9c0db011b5d256476b602b92fec92849fc2b36c"
        ],
        "threshold": 1
      }
    },
    "expiration": "2021-10-26T16:53:46Z",
    "metadata_spec_version": "0.6.0",
    "timestamp": "2020-10-26T16:53:46Z",
    "type": "root",
    "version": 1
  }
}

TEST_ROOT_MD_V2 = {
  "signatures": {
    "a59cea0987ee9046d68d2d011e919eb9278e3f478cca77f5204d65191ff8d7a5": {
      "other_headers": "04001608001d1621040a14b126c986f276831c7b04134f35b47db4364305025f96ff1b",
      "signature": "d406839499630a75350ba6f6c009aae173f15dd8c9be069c9b535ff77b6d624d6092487fe18e2c4f5c13b252a3ebe3f89ab15f4c52c66db752f8cbbfc6d96609"
    },
    "c8bd83b3bfc991face417d97b9c0db011b5d256476b602b92fec92849fc2b36c": {
      "other_headers": "04001608001d162104917adb684e2e9fb5ed4e59909ddd19a1268b62d005025f96ff1b",
      "signature": "f4c13d4456028778026639fcdc63ec7d6005e1e88f2dcfaf87afa3b89ce6a1ec8938af83fdc9d3d7045d0ebd648654c6af027daaf7164e014a8a53f373e9b906"
    }
  },
  "signed": {
    "delegations": {
      "key_mgr": {
        "pubkeys": [
          "013ddd714962866d12ba5bae273f14d48c89cf0773dee2dbf6d4561e521c83f7"
        ],
        "threshold": 1
      },
      "root": {
        "pubkeys": [
          "c8bd83b3bfc991face417d97b9c0db011b5d256476b602b92fec92849fc2b36c"
        ],
        "threshold": 1
      }
    },
    "expiration": "2021-10-26T16:53:46Z",
    "metadata_spec_version": "0.6.0",
    "timestamp": "2020-10-26T16:53:46Z",
    "type": "root",
    "version": 2
  }
}



# NOTE to dev:
#  test_authenticate was originally a long sequence of tests in a single
#  function.  I pulled out most of it, and what remains is has to be compared
#  to the new tests to see if it's still useful.
def test_wrap_sign_verify_signable():

    # Make a new keypair.  Returns keys and writes keys to disk.
    # Then load it from disk and compare that to the return value.  Exercise
    # some of the functions redundantly.
    generated_private, generated_public = gen_and_write_keys('keytest_new')
    loaded_new_private_bytes, loaded_new_public_bytes = keyfiles_to_bytes(
            'keytest_new')
    loaded_new_private, loaded_new_public = keyfiles_to_keys('keytest_new')

    old_private = PrivateKey.from_bytes(REG__PRIVATE_BYTES)
    old_public = PublicKey.from_bytes(REG__PUBLIC_BYTES)

    assert generated_private.is_equivalent_to(loaded_new_private)
    assert generated_public.is_equivalent_to(loaded_new_public)
    assert loaded_new_private.is_equivalent_to(
                PrivateKey.from_bytes(loaded_new_private_bytes))
    assert loaded_new_public.is_equivalent_to(
                PublicKey.from_bytes(loaded_new_public_bytes))


    # Clean up a bit for the next tests.
    new_private = loaded_new_private
    new_public = loaded_new_public
    del (
            loaded_new_public, loaded_new_private,
            generated_private, generated_public,
            loaded_new_private_bytes, loaded_new_public_bytes)




    # Test wrapping, signing signables, and verifying signables.
    d = {'foo': 'bar', '1': 2}
    d_modified = {'foo': 'DOOM', '1': 2}
    signable_d = wrap_as_signable(d)
    assert is_a_signable(signable_d)
    sign_signable(signable_d, old_private)
    assert is_a_signable(signable_d)

    verify_signable(
            signable=signable_d,
            authorized_pub_keys=[old_public.to_hex()],
            threshold=1)

    # Expect failure this time due to bad format.
    try:
        verify_signable(
                signable=signable_d['signed'],
                authorized_pub_keys=[old_public.to_hex()],
                threshold=1)
    except TypeError:
        pass
    else:
        assert False, 'Failed to raise expected exception.'

    # Expect failure this time due to non-matching signature.
    try:
        modified_signable_d = copy.deepcopy(signable_d)
        modified_signable_d['signed'] = d_modified
        verify_signable(
                signable=modified_signable_d,
                authorized_pub_keys=[old_public.to_hex()],
                threshold=1)
    except SignatureError:
        pass
    else:
        assert False, 'Failed to raise expected exception.'


    # Clean up a bit.
    for fname in [
            'keytest_new.pub', 'keytest_new.pri',
            'keytest_old.pri', 'keytest_old.pub']:
        if os.path.exists(fname):
            os.remove(fname)



# def test_repodata_verify_funcs():

#     # Test construction and verification of signed repodata_verify, including
#     # wrapping, signing the signable, and verifying the signables with a real
#     # example.
#     repodata_hashmap = {
#             "noarch/current_repodata.json":
#             "908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe",
#             "noarch/repodata.json":
#             "908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe",
#             "noarch/repodata_from_packages.json":
#             "908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe",
#             "osx-64/current_repodata.json":
#             "8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2",
#             "osx-64/repodata.json":
#             "8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2",
#             "osx-64/repodata_from_packages.json":
#             "8120fb07a6a8a280ffa2b89fb2fbb89484823d0b0357ff0cfa7c333352b2faa2"}

#     rd_v_md = build_repodata_verification_metadata(repodata_hashmap)
#     signable_rd_v_md = wrap_as_signable(rd_v_md)
#     assert is_a_signable(signable_rd_v_md)
#     sign_signable(signable_rd_v_md, old_private)
#     assert is_a_signable(signable_rd_v_md)

#     verify_signable(
#             signable=signable_rd_v_md,
#             authorized_pub_keys=[old_public.to_hex()],
#             threshold=1)

#     # Expect failure this time due to non-matching signature.
#     try:
#         modified_signable_rd_v_md = copy.deepcopy(signable_rd_v_md)

#         modified_signable_rd_v_md[
#                 'signed']['secured_files']['noarch/current_repodata.json'
#                 ] = modified_signable_rd_v_md['signed']['secured_files'][
#                 'noarch/current_repodata.json'][:-1] + 'f' # TODO: Generalize test condition. (Also, un-ugly.)

#         verify_signable(
#                 signable=modified_signable_rd_v_md,
#                 authorized_pub_keys=[old_public.to_hex()],
#                 threshold=1)
#     except SignatureError:
#         pass
#     else:
#         assert False, 'Failed to raise expected exception.'

#     # DEBUG: 💥💥💥💥 Dump the various bits and pieces for debugging.
#     #        Remove this.
#     with open('_test_output__repodata_hashmap.json', 'wb') as fobj:
#         fobj.write(canonserialize(repodata_hashmap))
#     with open('_test_output__repodata_verify.json', 'wb') as fobj:
#         fobj.write(canonserialize(signable_rd_v_md))


#     # Additional regression test for a file produced by the indexer.
#     # This should come up as good.
#     verify_signable(
#         signable={
#           "signatures": {
#             "013ddd714962866d12ba5bae273f14d48c89cf0773dee2dbf6d4561e521c83f7": "740a426113cb83a62e58eb41fcd0b5f36691b0b18bffbe7eb3da30b5baf83f6c703a0fdb584599702470c74f55572a27cf9de250fc3afb723c43fef4dc778401"
#           },
#           "signed": {
#             "expiration": "2019-10-28T15:36:32Z",
#             "metadata_spec_version": "0.0.4",
#             "secured_files": {
#               "noarch/current_repodata.json": "908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe",
#               "noarch/repodata.json": "908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe",
#               "noarch/repodata_from_packages.json": "908724926552827ab58dfc0bccba92426cec9f1f483883da3ff0d8664e18c0fe",
#               "osx-64/current_repodata.json": "fc9268ea2b4add37e090b7f2b2c88b95c513cab445fb099e8631d8815a384ae4",
#               "osx-64/repodata.json": "fc9268ea2b4add37e090b7f2b2c88b95c513cab445fb099e8631d8815a384ae4",
#               "osx-64/repodata_from_packages.json": "fc9268ea2b4add37e090b7f2b2c88b95c513cab445fb099e8631d8815a384ae4"
#             },
#             "timestamp": "2019-09-27T15:36:32Z",
#             "type": "repodata_verify"
#           }
#         },
#         authorized_pub_keys=[old_public.to_hex()],
#         threshold=1)


def test_sign_and_verify():
    """
    Tests functions:
        - sign
        - verify
    """

    # Generate new keys and construct key objects for old keys.
    new_private, new_public = gen_keys()
    old_private = PrivateKey.from_bytes(REG__PRIVATE_BYTES)
    old_public = PublicKey.from_bytes(REG__PUBLIC_BYTES)

    old_sig = old_private.sign(REG__MESSAGE_THAT_WAS_SIGNED)
    new_sig = new_private.sign(REG__MESSAGE_THAT_WAS_SIGNED)
    new_sig2 = new_private.sign(REG__MESSAGE_THAT_WAS_SIGNED)
    assert new_sig == new_sig2  # deterministic (obv not a thorough test)
    assert old_sig == REG__SIGNATURE # regression

    # Test verify()

    # Good signatures first.
    old_public.verify(REG__SIGNATURE, REG__MESSAGE_THAT_WAS_SIGNED)
    old_public.verify(old_sig, REG__MESSAGE_THAT_WAS_SIGNED)
    new_public.verify(new_sig, REG__MESSAGE_THAT_WAS_SIGNED)

    # Use wrong public key.
    wrong_pubkey_obj = PublicKey.from_hex(
            '0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef')
    with pytest.raises(cryptography.exceptions.InvalidSignature):
        wrong_pubkey_obj.verify(REG__SIGNATURE, REG__MESSAGE_THAT_WAS_SIGNED)

    # Use bad data.
    with pytest.raises(cryptography.exceptions.InvalidSignature):
        old_public.verify(new_sig, REG__MESSAGE_THAT_WAS_SIGNED + b'a')

    with pytest.raises(cryptography.exceptions.InvalidSignature):
        new_public.verify(new_sig, REG__MESSAGE_THAT_WAS_SIGNED[0:-1])



def test_verify_signature():
    verify_signature(
            REG__SIGNATURE_HEX,
            PublicKey.from_bytes(REG__PUBLIC_BYTES),
            REG__MESSAGE_THAT_WAS_SIGNED)

    # invalid signatures
    with pytest.raises(cryptography.exceptions.InvalidSignature):
        verify_signature(
                REG__SIGNATURE_HEX[:-6] + 'ffffff', # wrong value
                PublicKey.from_bytes(REG__PUBLIC_BYTES),
                REG__MESSAGE_THAT_WAS_SIGNED)

    with pytest.raises(TypeError):
        verify_signature(
                REG__SIGNATURE_HEX[:-1], # wrong length
                PublicKey.from_bytes(REG__PUBLIC_BYTES),
                REG__MESSAGE_THAT_WAS_SIGNED)

    with pytest.raises(TypeError):
        verify_signature(
                REG__SIGNATURE, # wrong type
                PublicKey.from_bytes(REG__PUBLIC_BYTES),
                REG__MESSAGE_THAT_WAS_SIGNED)


    # other bad args
    with pytest.raises(cryptography.exceptions.InvalidSignature):
        verify_signature(
                REG__SIGNATURE_HEX, # wrong type
                PublicKey.from_bytes(REG__PUBLIC_BYTES),
                REG__MESSAGE_THAT_WAS_SIGNED + b'\xc9') # altered message

    with pytest.raises(cryptography.exceptions.InvalidSignature):
        verify_signature(
                REG__SIGNATURE_HEX,
                PublicKey.from_bytes(REG__PUBLIC_BYTES[:-4] + b'0000'), # wrong key
                REG__MESSAGE_THAT_WAS_SIGNED)

    with pytest.raises(TypeError):
        verify_signature(
                REG__SIGNATURE_HEX,
                REG__PUBLIC_BYTES,                  # wrong type
                REG__MESSAGE_THAT_WAS_SIGNED)

    with pytest.raises(TypeError):
        verify_signature(
                REG__SIGNATURE_HEX,
                PublicKey.from_bytes(REG__PUBLIC_BYTES),
                {'this is not bytes': 1})              # wrong type


# verify_root is also tested in test_root.py (but test_root.py expects GPG)
def test_verify_root():
    """
    Tests conda_content_trust.authentication.verify_root
    """

    # Root chaining: normal test
    verify_root(TEST_ROOT_MD_V1, TEST_ROOT_MD_V2)


    # Now we tinker a bit to break stuff.
    root_v2_edited = copy.deepcopy(TEST_ROOT_MD_V2)


    # Can't verify root v10 using root v1 (chaining)
    with pytest.raises(MetadataVerificationError):
        root_v2_edited['signed']['version'] = 10
        verify_root(TEST_ROOT_MD_V1, root_v2_edited)

    # Reset.
    root_v2_edited['signed']['version'] = TEST_ROOT_MD_V2['signed']['version']


    # Bad signature, same keys, same contents
    # with pytest.raises(cryptography.exceptions.InvalidSignature):
    with pytest.raises(SignatureError):
        sig = root_v2_edited['signatures'][REG__PUBLIC_HEX_ROOT]['signature']
        sig = sig[:-6] + 'ffffff'
        root_v2_edited['signatures'][REG__PUBLIC_HEX_ROOT]['signature'] = sig
        verify_root(TEST_ROOT_MD_V1, root_v2_edited)

    # Reset.
    root_v2_edited['signatures'] = copy.deepcopy(TEST_ROOT_MD_V2['signatures'])


    # Not enough signatures from authorized keys:
    #     Have one of the signatures claim to be from the wrong key.
    with pytest.raises(SignatureError):
        root_v2_edited['signatures'][REG__PUBLIC_HEX_ROOT[:-6] + 'ffffff'] \
                = root_v2_edited['signatures'][REG__PUBLIC_HEX_ROOT]
        del root_v2_edited['signatures'][REG__PUBLIC_HEX_ROOT]
        verify_root(TEST_ROOT_MD_V1, root_v2_edited)

    # Reset.
    root_v2_edited['signatures'] = copy.deepcopy(TEST_ROOT_MD_V2['signatures'])


    # Not enough signatures from authorized keys:
    #     Change the trusted metadata such that we expect sigs from 3 distinct
    #     authorized keys (and still provide only 2).
    with pytest.raises(SignatureError):
        root_v1_edited = copy.deepcopy(TEST_ROOT_MD_V1)
        root_v1_edited['signed']['delegations']['root']['threshold'] += 1
        verify_root(root_v1_edited, TEST_ROOT_MD_V2)

    # Reset.
    root_v1_edited['signed']['delegations']['root']['threshold'] -= 1



# def test_verify_delegation():
#     """
#     Tests conda_content_trust.authentication.verify_delegation
#     """
#     raise NotImplementedError('verify_delegation requires unit tests.')



# -*- coding: utf-8 -*-

# # If nacl/encryption is optional.  Also, the checks in functions would probably
# # then be best as a wrapper.
# try:
#     import nacl.utils
#     import nacl.public
#     NACL_AVAILABLE = True

# except ImportError:
#     print(
#         'Importing car.encrypt, but note that the functions in this module '
#         'require  PyNaCl, which appears not to be available.')
#     NACL_AVAILABLE = False


def generate_nacl_keypair():
    """
    Returns two objects, (a nacl.public.PublicKey, a nacl.public.PrivateKey)
    for encryption using nacl.public.Box or nacl.public.SealedBox.
    (We'll use SealedBox in this code, since we provide code signing
    separately.)
    """
    # if not NACL_AVAILABLE:
    #     # TODO: Consider a dependency error class.
    #     raise Exception('encrypt() requires PyNaCl, which is not available.')

    private = nacl.public.PrivateKey.generate()
    public = private.public_key

    return private, public


def raw_key_from_nacl_key(nacl_key):
    """
    Given an nacl.public.PrivateKey or an nacl.public.PublicKey, returns the
    underlying 32-byte key value (a Curve25519 key).
    """
    if not (isinstance(nacl_key, nacl.public.PublicKey)
            or isinstance(nacl_key, nacl.public.PrivateKey)):
        raise TypeError('Expecting a PublicKey or PrivateKey object')
    return nacl_key.encode()


def private_nacl_key_from_raw_key(key_value):
    """
    Given a raw key value (32 bytes, a Curve25519 key), generates an
    nacl.public.PrivateKey object using that key value, which can then be used
    for encryption and decryption via the PyNaCl library (nacl.public).
    """
    # ✅ TODO: Argument validation (bytes, Python2 & Python3)
    nacl.public.PrivateKey(key_value)


def public_nacl_key_from_raw_key(key_value):
    """
    Given a raw key value (32 bytes, a Curve25519 key), generates an
    nacl.public.PublicKey object using that key value, which can then be used
    for encryption via the PyNaCl library (nacl.public).
    """
    # ✅ TODO: Argument validation (bytes, Python2 & Python3)
    nacl.public.PublicKey(key_value)



def encrypt(data, public_key):
    """
    Takes bytes and a public key (nacl.public.PublicKey), returns encrypted
    data (bytes) that can be decrypted using the corresponding private key.
    """
    # if not NACL_AVAILABLE:
    #     # TODO: Consider a dependency error class.
    #     raise Exception('encrypt() requires PyNaCl, which is not available.')

    if not isinstance(public_key, nacl.public.PublicKey):
        raise TypeError('Arg "public_key" must be a nacl.public.PublicKey.')

    encryptor = nacl.public.SealedBox(public_key)

    return encryptor.encrypt(data)



def decrypt(encrypted, private_key):
    """
    Takes bytes and a private key (nacl.public.PrivateKey), returns decrypted
    data (bytes).
    """
    # if not NACL_AVAILABLE:
    #     # TODO: Consider a dependency error class.
    #     raise Exception('decrypt() requires PyNaCl, which is not available.')

    if not isinstance(private_key, nacl.public.PrivateKey):
        raise TypeError('Arg "private_key" must be a nacl.public.PrivateKey.')

    decryptor = nacl.public.SealedBox(private_key)

    return decryptor.decrypt(encrypted)
